/*
 * MGL -- MobileGear Graphic Library -
 * Copyright (C) 1998, 1999
 *      Koji Suzuki (suz@at.sakura.ne.jp)
 *      Yukihiko Sano (yukihiko@yk.rim.or.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY KOJI SUZUKI AND YUKIHIKO SANO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
const static int same_format=0;
#define MD_PUT_SCANSEGMENT(dst_y,dst_x,src,nbytes)

const static int show_mouse=1;
#define TOUCH_DEVICE            "/dev/tpanel"
static int                      fd;
static	int x, y, z, b, flags;
/*------------------------------------------------------------------------------*/

/*
 * $Id$
 *
 * Touch-panel driver
 *
 * Designed for for use with the Linux VR project touch-panel kernel driver.
 * http://linux-vr.org
 *
 * Copyright 2000 Bradley D. LaRonde <brad@ltc.com>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Bradley D. LaRonde not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Bradley D. LaRonde makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * BRADLEY D. LARONDE DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL BRADLEY D. LARONDE
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <sys/ioctl.h>
#include <linux/tpanel.h>
#include "tptrans.h"
*/

#define KD_BUTTON_1     0x01
#define KD_BUTTON_2     0x02
#define KD_BUTTON_3     0x04
#define KD_MOUSE_DELTA  0x80000000

#define	MIN(a,b) (((a)<(b))?(a):(b))

/*
 * Enable absolute coordinate transformation.
 * Normally this should be left at 1.
 * To disable transformation, set it to 0 before calling VrPanelInit.
 * This is done by the pointer calibration utility since it's used to produce the pointer
 * calibration data file.
 */
int Vr_enable_transform = 1;

static TRANSFORMATION_COEFFICIENTS tc;

int GetPointerCalibrationData()
{
	/*
	 * Read the calibration data from the calibration file.
	 * Calibration file format is seven coefficients separated by spaces.
	 */

	/* Get pointer calibration data from this file */
	const char cal_filename[] = "/etc/pointercal";

	int items;

	FILE* cf = fopen(cal_filename, "r");
	if ( cf == NULL )
	{
		fprintf(stderr, "Error %d opening pointer calibration file %s.\n",
			errno, cal_filename);
		return -1;
	}

	items = fscanf(cf, "%d %d %d %d %d %d %d",
		&tc.a, &tc.b, &tc.c, &tc.d, &tc.e, &tc.f, &tc.s);
	if ( items != 7 )
	{
		fprintf(stderr, "Improperly formatted pointer calibration file %s.\n",
			cal_filename);
		fclose(cf);
		return -1;
	}

#if TEST
		printf("a=%d b=%d c=%d d=%d e=%d f=%d s=%d\n",
			tc.a, tc.b, tc.c, tc.d, tc.e, tc.f, tc.s);
#endif

	fclose(cf);
	return 0;
}

typedef struct
{
	int x, y;
} XYPOINT;

inline XYPOINT DeviceToScreen(XYPOINT p)
{
	/*
	 * Transform device coordinates to screen coordinates.
	 * Take a point p in device coordinates and return the corresponding
	 * point in fractional screen coodinates (fraction is
	 * 1 / TRANSFORMATION_UNITS_PER_PIXEL).
	 *
	 * It can scale, translate, rotate and/or skew, based on the coefficients
	 * calculated above based on the list of screen vs. device coordinates.
	 */

	static XYPOINT prev;
	/* Slop is how far the pointer has to move before we will allow the output
	 * to change.  Optimal is probably 3/4 of a pixel, but that might not be
	 * enough for some panels that are very noisey.
	 * Using larger values appears to makes the pointer movement more "blocky".
	 * This is noticeable, say,  when drawing a line at a 45 degree angle.
	 */
	/* TODO: make this configurable */
	const short slop = TRANSFORMATION_UNITS_PER_PIXEL * 3 / 4;
	XYPOINT new, out;

	/* transform */
	new.x = (tc.a * p.x + tc.b * p.y + tc.c) / tc.s;
	new.y = (tc.d * p.x + tc.e * p.y + tc.f) / tc.s;

	/* hysteresis (thanks to John Siau) */
	if ( abs(new.x - prev.x) >= slop )
		out.x = (new.x | 0x3) ^ 0x3;
	else
		out.x = prev.x;

	if ( abs(new.y - prev.y) >= slop )
		out.y = (new.y | 0x3) ^ 0x3;
	else
		out.y = prev.y;

	prev = out;

	return out;
}

static int VrTpanelInit()
{
 	/*
	 * Open up the touch-panel device.
	 * Return the fd if successful, or negative if unsuccessful.
	 */

	struct scanparam s;
	int result;

	int fd = open(TOUCH_DEVICE, O_NONBLOCK);
	if (fd < 0) {
		fprintf(stderr, "Error %d opening touch panel\n", errno);
		return -1;
	}

	/*
	 * Set sample interval in microseconds to get s sample frequency of 50Hz.
	 * If you adjust this value, you may also want to adust iir_shift_bits
	 * in VrTpanelRead.
	 */
	s.interval = 20000;

	/*
	 * Functional upper limit on settle time is approximately (scan_interval / 5) - 60
	 * (5 conversions and some fixed overhead fitting within scan interval).
	 * The opmtimal value is the lowest that doesn't cause significant distortion.
	 * The lower the value, the more "simultaneous" and thus coherent the conversions.
	 * 480us is probably a reasonable value for most hardware.  Lower than that can
	 * get into distortion.
	 */
	s.settletime = 480;

	result = ioctl(fd, TPSETSCANPARM, &s);
	if ( result < 0 )
		fprintf(stderr, "Error %d, result %d setting scan parameters.\n", result, errno);

	if (Vr_enable_transform)
	{ 
		if (GetPointerCalibrationData() < 0)
		{
			close(fd);
			return -1;
		}
	}

	return fd;
}

/* Mouse button bits in BUTTON*/
#define LBUTTON		04
#define MBUTTON		02
#define RBUTTON		01

static int VrTpanelRead(int *px, int *py, int *pz, unsigned int *pb)
{
	/*
	 * This driver returns absolute postions, not deltas.
         * Returns the position in px, py, pressure in pz, and buttons pb.
	 * Returns -1 on error.
	 * Returns 0 if no new data is available.
         * Does not return 1 (which is for relative position data).
	 * Returns 2 if position data is absolute (i.e. touch panels).
	 * Returns 3 if position data is not available, but button data is.
	 * This routine does not block.
	 */

	/* If z is below this value, ignore the data. */
	/* TODO: make this configurable */
	const int low_z_limit = 10 ;/* default is 900 */

	/*
	 * I do some error masking by tossing out really wild data points.
	 * Lower data_change_limit value means pointer get's "left behind" more easily.
	 * Higher value means less errors caught.
	 * The right setting of this value is just slightly higher than the number of
	 * units traversed per sample during a "quick" stroke.
	 * I figure a fast pen scribble can cover about 3000 pixels in one second on a typical screen.
	 * There are typically about 3 to 6 points of touch-panel resolution per pixel.
	 * So 3000 pixels-per-second * 3 to 6 tp-points-per-pixel / 50 samples-per-second =
	 * 180 to 360 tp points per scan.
	 */
	/* TODO: make this configurable */
	const int data_change_limit = 360;
	static int have_last_data = 0;
	static int last_data_x = 0;
	static int last_data_y = 0;

	/*
	 * Thanks to John Siau <jsiau@benchmarkmedia.com> for help with the noise filter.
	 * I use an infinite impulse response low-pass filter on the data to filter out
	 * high-frequency noise.  Results look better than a finite impulse response filter.
	 * If I understand it right, the nice thing is that the noise now acts as a dither
	 * signal that effectively increases the resolution of the a/d converter by a few bits
	 * and drops the noise level by about 10db.
	 * Please don't quote me on those db numbers however. :-)
	 * The end result is that the pointer goes from very fuzzy to much more steady.
	 * Hysteresis really calms it down in the end (elsewhere).
	 *
	 * iir_shift_bits effectively sets the number of samples used by the filter
	 * (number of samples is 2^iir_shift_bits).
	 *
	 * Setting iir_shift_bits lower allows shorter "taps" and less pointer lag, but a
	 * fuzzier pointer, which can be especially bad for display update when dragging.
	 *
	 * Setting iir_shift_bits higher requires longer "presses" and causes more pointer lag,
	 * but it provides steadier pointer.
	 *
	 * If you adjust iir_shift_bits, you may also want to adjust the sample interval
	 * in VrTpanelInit. 
	 *
	 * The filter gain is fixed at 8 times the input (gives room for increased resolution
	 * potentially added by the noise-dithering).
	 *
	 * The filter won't start outputing data until iir_count is above iir_output_threshold.
	 */
	const int iir_shift_bits = 2;
	const int iir_sample_depth = (1 << iir_shift_bits);
	const int iir_output_threshold = 1;
	const int iir_gain = 8;
	static int iir_accum_x = 0;
	static int iir_accum_y = 0;
	static int iir_accum_z = 0;
	static int iir_count = 0;
	int iir_out_x;
	int iir_out_y;
	int iir_out_z;

	int data_x, data_y, data_z;

	/* read a data point */
	short data[6];
	int bytes_read;
	bytes_read = read(fd, data, sizeof(data));
	if (bytes_read != sizeof(data)) {
		if (errno == EINTR || errno == EAGAIN)
			return 0;
		return -1;
	}

	/* did we lose any data? */
	if ( (data[0] & 0x2000) )
		fprintf(stderr, "Lost touch panel data\n");

	/* do we only have contact state data (no position data)? */
	if ( (data[0] & 0x8000) == 0 )
	{
		/* is it a pen-release? */
		if ( (data[0] & 0x4000) == 0 )
		{
			/* TODO: in order to provide maximum responsiveness
			 * maybe return the data in the filter
			 * if the filter count is still below
			 * the output threshold.
			 */

			/* reset the limiter */
			have_last_data = 0;

			/* reset the filter */
			iir_count = 0;
			iir_accum_x = 0;
			iir_accum_y = 0;
			iir_accum_z = 0;

			/* return the pen (button) state only, */
			/* indicating that the pen is up (no buttons are down) */
			*pb = 0;
			return 3;
		}

		/* ignore pen-down since we don't know where it is */
		return 0;
	}

	/* we have position data */

	/*
	 * Combine the complementary panel readings into one value (except z)
	 * This effectively doubles the sampling freqency, reducing noise by approx 3db.
	 * Again, please don't quote the 3db figure.
	 * I think it also cancels out changes in the overall resistance of the panel
	 * such as may be caused by changes in panel temperature.
	 */
	data_x = data[2] - data[1];
	data_y = data[4] - data[3];
	data_z = data[5];

	/* isn't z big enough for valid position data? */
	if ( data_z <= low_z_limit )
		return 0;

	/* has the position changed more than we will allow? */
	if ( have_last_data )
		if ( (abs(data_x - last_data_x) > data_change_limit)
			|| ( abs(data_y - last_data_y) > data_change_limit ) )
			return 0;

	/* save last position */
	last_data_x = data_x;
	last_data_y = data_y;
	have_last_data = 1;

	/* is filter full? */
	if ( iir_count == iir_sample_depth )
	{
		/* make room for new sample */
		iir_accum_x -= iir_accum_x >> iir_shift_bits;
		iir_accum_y -= iir_accum_y >> iir_shift_bits;
		iir_accum_z -= iir_accum_z >> iir_shift_bits;
		iir_count--;
	}

	/* feed new sample to filter */
	iir_accum_x += data_x;
	iir_accum_y += data_y;
	iir_accum_z += data_z;
	/* TODO: optimize the unnecessary count--/++ when filter full */
	iir_count++;

	/* aren't we over the threshold yet? */
	if (iir_count <= iir_output_threshold)
		return 0;

	/* Touch-panel only reports left button. */
	/* I moved this up here in preparation for the below-threshold
	 * on pen-up filter output TODO above.
	 */
	*pb = LBUTTON;

	/* figure filter output */
	/* TODO: optimize for shifts instead of divides (when possible)? */
	iir_out_x = (iir_accum_x * iir_gain) / iir_count;
	iir_out_y = (iir_accum_y * iir_gain) / iir_count;
	iir_out_z = (iir_accum_z * iir_gain) / iir_count;

	/* transformation enabled? */
	if (Vr_enable_transform)
	{
		/* transform filter output to screen coordinates */
		XYPOINT transformed = {iir_out_x, iir_out_y};
		transformed = DeviceToScreen(transformed);

		/*
		 * HACK: move this from quarter pixels to whole pixels for now
		 * at least until I decide on the right interface to get the
		 * quarter-pixel data up to the next layer.
		 */
		*px = transformed.x >> 2;
		*py = transformed.y >> 2;
	}
	else
	{
		/* return untransformed coords */
		*px = iir_out_x;
		*py = iir_out_y;
	}

	/* Return filtered pressure.*/
	*pz = iir_out_z;
	
	return 2;
}



static int open_touch(void)
{
	fd = VrTpanelInit();
	fprintf(stderr, "vr41xx touch-panel fd %d\n", fd);
	if ( 0 > fd )
		return -1;
	return 0;
}

static int MD_MOUSE_EVENT(void)
{
/*	int x, y, z, b, flags; */
	int result = VrTpanelRead(&x, &y, &z, &b);
        if ( result <= 0 )
                return 0;
        return 1;

}

static void close_touch(void)
{
	if(fd >= 0)
		close(fd);
}


/*-------------------------------------------------------------------------------*/
static void MD_SET_MOUSEXY(int xs,int ys) {
	x = xs;
	y = ys;
}

static inline int MD_MOUSE_X() { return x; }
static inline int MD_MOUSE_Y() { return y; }

static int old_z = 0;
static inline int MD_MOUSE_BUTTON() {
	if ( z - old_z > 10 ){
	old_z = z;
	return 1;
	}else{
	old_z = z;
	return 0;
	}
}
